##for random_pattern and tang dataset
import math
import torch
import torch.nn as nn
from config import args
import torch.nn.functional as F
import numpy as np
import random
from torch.utils.data import DataLoader
from model_T import *
np.random.seed(args.seed2)
random.seed(args.seed2)
torch.manual_seed(args.seed1)
torch.cuda.manual_seed_all(args.seed1)

class TransformerModel(nn.Module):

    def __init__(self, ntoken, ninp, nhead, nhid, nlayers, dropout=0.5):
        super(TransformerModel, self).__init__()
        from torch.nn import TransformerEncoder, TransformerEncoderLayer
        self.model_type = 'Transformer'
        self.pos_encoder = PositionalEncoding(ninp, dropout)
        encoder_layers = TransformerEncoderLayer(ninp, nhead, nhid, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.ninp = ninp
        self.decoder = nn.Linear(ninp, ntoken)

        self.init_weights()

    def generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, src, src_mask):
        src = self.encoder(src) * math.sqrt(self.ninp)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, src_mask)
        output = self.decoder(output)
        return output
class PositionalEncoding(nn.Module):

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)

device = torch.device("cpu" if args.cuda < 0 else "cuda:" + str(args.cuda))
def data_gen_all(V,num_samples):
    data=torch.from_numpy(np.random.randint(1,V,size=(num_samples,args.len_seq)))
    data[:,0]=0
    print(data)
    return data
   
def make_mask(tgt,pad=0): 
    tgt_mask=(tgt!=pad).unsqueeze(-2)
    tgt_mask=tgt_mask & subsequent_mask(tgt.size(-1)).type_as(tgt_mask.data)
    return tgt_mask   
if args.data_type=='random_pattern':
    train_data = data_gen_all(args.V,args.num_samples)
    val_data = data_gen_all(args.V, args.num_samples)
    #val_data=train_data.t()
    test_data = data_gen_all(args.V,args.num_samples)
elif args.data_type=='tang':
    datas = np.load("tang.npz",allow_pickle=True)
    data = datas['data']
    print(data.shape[0])
    print(np.max(data))
    ix2word = datas['ix2word'].item()
    #print(ix2word)
    word2ix = datas['word2ix'].item()
    #print(word2ix)
    print(ix2word[data[1][0]])
    train_data = torch.from_numpy(data)[:args.num_train].long()
    test_data=torch.from_numpy(data)[:args.num_train].long()
    val_data= torch.from_numpy(data)[:args.num_train].long()  
train_dataloader=DataLoader(train_data,batch_size=args.batch_size,
                            shuffle=True,
                            num_workers=2)  
val_dataloader=DataLoader(val_data,batch_size=args.batch_size,
                            shuffle=True,
                            num_workers=2)   
test_dataloader=DataLoader(test_data,batch_size=args.batch_size,
                            shuffle=True,
                            num_workers=2)

dropout = 0.2 # the dropout value
if args.data_type=='random_pattern':
    ntokens = args.V
else:
    ntokens = args.vocab_size    
    print(ntokens)
if args.model_type=='gru':
    from model import *
    model= PoetryModel(args)
    '''
    if args.model_path:
        model.load_state_dict(t.load(args.model_path,map_location='cpu'))
    '''
elif args.model_type=='mytransf':
    model=make_model(ntokens,N=args.N,d_model=args.d_model,d_ff=args.d_ff,h=args.h,dropout=0.2)
elif args.model_type=='pytransf':
    model=TransformerModel(ntokens, args.d_model, args.h, args.d_ff, args.N, dropout)
model.to(device)
#model = TransformerModel(args.V, args.d_model, args.h, args.d_ff, args.N, dropout).to(device)
criterion = nn.CrossEntropyLoss()
lr = args.lr # learning rate
if args.optimizer=='adam':
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
elif args.optimizer=='sgd':
    optimizer =torch.optim.SGD(model.parameters(), lr=lr)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, 1, gamma=0.95)

import time
def train():
    model.train() # Turn on the train mode
    total_loss = 0.
    start_time = time.time()
    all_samples=0
    #for batch, i in enumerate(range(0,  bptt-1, bptt)):
    for i, batch in (enumerate(train_dataloader)):
        #data, targets = get_batch(train_data, i)
        bptt=batch.size(1)
        now_samples=batch.size(0)
        all_samples+=batch.size(0)
        optimizer.zero_grad()
        if args.model_type=='gru':
            batch=batch.transpose(1,0).to(args.device)
            data,targets=batch[:-1],batch[1:]
            targets=targets.reshape(-1)
            output,_ = model(data)
            loss = criterion(output, targets)
        elif args.model_type=='mytransf':
            batch=batch.to(args.device)
            data,targets=batch[:,:-1],batch[:,1:]
            targets=targets.reshape(-1)
            src_mask=make_mask(data,-1).to(args.device)
            output = model(data, src_mask)
            loss = criterion(output.view(-1, ntokens), targets)
        else:
            batch=batch.transpose(1,0).to(args.device)
            data,targets=batch[:-1],batch[1:]
            targets=targets.reshape(-1)
            src_mask = model.generate_square_subsequent_mask(data.size(0)).to(args.device)
            output = model(data, src_mask)
            loss = criterion(output.view(-1, ntokens), targets)
        loss.backward()
        torch.nn.utils.clip_grad_norm_(model.parameters(), 0.5)
        optimizer.step()

        total_loss += loss.item()*now_samples
        log_interval = 200
        if i%log_interval==0:
            cur_loss = total_loss/all_samples
            elapsed = time.time() - start_time
            print('| epoch {:3d} |  '
                  'lr {:02.2f} | ms/batch {:5.2f} | '
                  'loss {:.8g} | ppl {:.8g}'.format(
                    epoch, scheduler.get_lr()[0],
                    elapsed * 1000 / log_interval,
                    cur_loss, math.exp(cur_loss)))
            #total_loss = 0
            start_time = time.time()
    print(all_samples)
    print(total_loss/all_samples)
def evaluate(eval_model, dataloader):
    eval_model.eval() # Turn on the evaluation mode
    total_loss = 0.
    
    #src_mask = model.generate_square_subsequent_mask(args.len_seq-1).to(device)
    with torch.no_grad():
        for i, batch in (enumerate(dataloader)):
            #data, targets = get_batch(data_source, i)
            #batch=batch.transpose(1,0).to(args.device)
            if args.model_type=='gru':
                batch=batch.transpose(1,0).to(args.device)
                data,targets=batch[:-1],batch[1:]
                targets=targets.reshape(-1)
                output,_ = eval_model(data)
                loss = criterion(output, targets)
            elif args.model_type=='mytransf':
                batch=batch.to(args.device)
                data,targets=batch[:,:-1],batch[:,1:]
                targets=targets.reshape(-1)
                src_mask=make_mask(data,-1).to(args.device)
                output = eval_model(data, src_mask)
                loss = criterion(output.view(-1, ntokens), targets)
            else:
                batch=batch.transpose(1,0).to(args.device)
                data,targets=batch[:-1],batch[1:]
                targets=targets.reshape(-1)
                src_mask = model.generate_square_subsequent_mask(data.size(0)).to(args.device)
                output = eval_model(data, src_mask)
                loss = criterion(output.view(-1, ntokens), targets)
         
            total_loss += batch.size(0) * loss.item()
            #print(total_loss)
    return total_loss / (i+1)
best_val_loss = float("inf")
epochs = 200 # The number of epochs
best_model = None

for epoch in range(1, epochs + 1):
    epoch_start_time = time.time()
    train()
    '''
    val_loss = evaluate(model, val_dataloader)
    #print('-' * 89)
    #print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | '
          #'valid ppl {:8.2f}'.format(epoch, (time.time() - epoch_start_time),
                                     #val_loss, math.exp(val_loss)))
    #print('-' * 89)

    if val_loss < best_val_loss:
        best_val_loss = val_loss
        best_model = model
    '''
    scheduler.step()
test_loss = evaluate(best_model, test_dataloader)
print('=' * 89)
print('| End of training | test loss {:5.2f} | test ppl {:8.2f}'.format(
    test_loss, math.exp(test_loss)))
print('=' * 89)