#for wikitext dataset
import math
import torch
import torch.nn as nn
from config import args
import torch.nn.functional as F
import numpy as np
import random
from torch.utils.data import DataLoader
from model_T import *
np.random.seed(args.seed2)
random.seed(args.seed2)
torch.manual_seed(args.seed1)
torch.cuda.manual_seed_all(args.seed1)
class TransformerModel(nn.Module):

    def __init__(self, ntoken, ninp, nhead, nhid, nlayers, dropout=0.5):
        super(TransformerModel, self).__init__()
        from torch.nn import TransformerEncoder, TransformerEncoderLayer
        self.model_type = 'Transformer'
        self.pos_encoder = PositionalEncoding(ninp, dropout)
        encoder_layers = TransformerEncoderLayer(ninp, nhead, nhid, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.ninp = ninp
        self.decoder = nn.Linear(ninp, ntoken)
        self.init_weights()
    def generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask
    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, src, src_mask):
        src = self.encoder(src) * math.sqrt(self.ninp)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, src_mask)
        output = self.decoder(output)
        return output
class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)
    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)
import torchtext
from torchtext.data.utils import get_tokenizer
TEXT = torchtext.data.Field(tokenize=get_tokenizer("basic_english"),
                            init_token='<sos>',
                            eos_token='<eos>',
                            lower=True)
train_txt, val_txt, test_txt = torchtext.datasets.WikiText2.splits(TEXT)
TEXT.build_vocab(train_txt)
#device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
device = torch.device("cpu" if args.cuda < 0 else "cuda:" + str(args.cuda))
def batchify(data, bsz):
    data = TEXT.numericalize([data.examples[0].text])
    # Divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    return data.to(device)

batch_size = args.batch_size
eval_batch_size = 10
train_data = batchify(train_txt, batch_size)
val_data = batchify(val_txt, eval_batch_size)
test_data = batchify(test_txt, eval_batch_size)
bptt = args.len_seq
#bptt=35
def get_batch(source, i):
    seq_len = min(bptt, len(source) - 1 - i)
    data = source[i:i+seq_len]
    target = source[i+1:i+1+seq_len]
    return data, target
ntokens = len(TEXT.vocab.stoi) # the size of vocabulary
if args.mpo:
   ntokens=28800
print(ntokens)
emsize = args.d_model # embedding dimension
nhid = args.d_ff # the dimension of the feedforward network model in nn.TransformerEncoder
nlayers = args.N # the number of nn.TransformerEncoderLayer in nn.TransformerEncoder
nhead = args.h # the number of heads in the multiheadattention models
dropout = 0.2 # the dropout value
def make_mask(tgt,pad=0): 
    tgt_mask=(tgt!=pad).unsqueeze(-2)
    tgt_mask=tgt_mask & subsequent_mask(tgt.size(-1)).type_as(tgt_mask.data)
    return tgt_mask 
if args.model_type=='pytransf':
   model = TransformerModel(ntokens, emsize, nhead, nhid, nlayers, dropout).to(device)
elif args.model_type=='mytransf':
   model=make_model(ntokens,N=args.N,d_model=args.d_model,d_ff=args.d_ff,h=args.h,dropout=0.2).to(device)
elif args.model_type=='gru':
   from model import *
   args.vocab_size=ntokens
   model= PoetryModel(args).to(device)
criterion = nn.CrossEntropyLoss()
lr = 5.0 # learning rate
if args.optimizer=='adam':
   optimizer = torch.optim.SGD(model.parameters(), lr=lr)
elif args.optimizer=='sgd':
   optimizer = torch.optim.SGD(model.parameters(), lr=lr)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, 1, gamma=0.95)

import time
def train():
    model.train() # Turn on the train mode
    total_loss = 0.
    start_time = time.time()
    #ntokens = len(TEXT.vocab.stoi)
    all_bits=0
    if args.model_type=='pytransf':
        src_mask = model.generate_square_subsequent_mask(bptt).to(device)
    for batch, i in enumerate(range(0,  train_data.size(0)-1, bptt)):
        data, targets = get_batch(train_data, i)
        bits_batch=data.size(0)*data.size(1)
        all_bits+=bits_batch
        optimizer.zero_grad()
        if args.model_type=='pytransf':
            if data.size(0) != bptt:
                src_mask = model.generate_square_subsequent_mask(data.size(0)).to(device)
            output = model(data,src_mask)
        elif args.model_type == 'mytransf':
            data,targets = data.t(),targets.t()
            src_mask= make_mask(data,pad=-1)
            output = model(data,src_mask)
        elif args.model_type=='gru':
            output,_=model(data)
        loss = criterion(output.reshape(-1, ntokens), targets.reshape(-1))
        loss.backward()
        torch.nn.utils.clip_grad_norm_(model.parameters(), 0.5)
        optimizer.step()

        total_loss += loss.item()*bits_batch
        log_interval = 200
        if batch % log_interval==0:
            cur_loss = total_loss / all_bits
            elapsed = time.time() - start_time
            print('| epoch {:3d} | {:5d}/{:5d} batches | '
                  'lr {:02.2f} | ms/batch {:5.2f} | '
                  'loss {:.8g} | ppl {:.8g}'.format(
                    epoch, batch, len(train_data) // bptt, scheduler.get_lr()[0],
                    elapsed * 1000 / log_interval,
                    cur_loss, math.exp(cur_loss)))
            #total_loss = 0
            start_time = time.time()
    print(total_loss/all_bits)
def evaluate(eval_model, data_source):
    eval_model.eval() # Turn on the evaluation mode
    total_loss = 0.
    #ntokens = len(TEXT.vocab.stoi)
    all_bits=0
    #if args.model_type=='pytransf':
        #src_mask = model.generate_square_subsequent_mask(bptt).to(device)
    with torch.no_grad():
        for i in range(0, data_source.size(0) - 1, bptt):
            data, targets = get_batch(data_source, i)
            all_bits+=data.size(0)*data.size(1)
            if args.model_type=='pytransf':
                #if data.size(0) != bptt:
                src_mask = model.generate_square_subsequent_mask(data.size(0)).to(device)
                output = eval_model(data, src_mask)
            elif args.model_type == 'mytransf':
                data,targets = data.t(),targets.t()
                src_mask= make_mask(data,pad=-1)
                output = eval_model(data, src_mask)
            elif args.model_type == 'gru':
                output,_ =model(data)
            output_flat = output.reshape(-1, ntokens)
            total_loss += data.size(0)*data.size(1) * criterion(output_flat, targets.reshape(-1)).item()
    print(all_bits)
    return total_loss / all_bits
best_val_loss = float("inf")
epochs = args.epoch # The number of epochs
best_model = model

for epoch in range(1, epochs + 1):
    
    train()
    epoch_start_time = time.time()
    val_loss = evaluate(model, val_data)
    print('-' * 89)
    print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | '
          'valid ppl {:8.2f}'.format(epoch, (time.time() - epoch_start_time),
                                     val_loss, math.exp(val_loss)))
    print('-' * 89)

    if val_loss < best_val_loss:
        best_val_loss = val_loss
        best_model = model
    #generation
    
    scheduler.step()
def generation(best_model,max_len=500,start=1):
    best_model.eval()
    generated_tokens=[]
    generated_tokens.append(start)
    with torch.no_grad():
        for i in range(max_len):
            input = torch.Tensor(generated_tokens).to(args.device).view(-1,1).long() 
            #print(input.shape)
            src_mask=None
            if args.model_type=='mytransf':
                src_mask= make_mask(input.t(),pad=-1)
                output=model(input.t(),src_mask)
                #print(output.shape)
                top_index = output[-1,:].topk(1)[1][0].item()
                generated_tokens.append(top_index)
    return generated_tokens      
if args.test:       
    test_loss = evaluate(best_model, test_data)
    print('=' * 89)
    print('| End of training | test loss {:5.2f} | test ppl {:8.2f}'.format(
        test_loss, math.exp(test_loss)))
    print('=' * 89)
print('begin generation')
time0=time.time()
print(time.time()-time0)
generated_tokens=generation(best_model,max_len=args.max_gen_len)
print('used time for generation:',time.time()-time0)
#print(generated_tokens)
print(len(generated_tokens))