from main import *
from config import args
import torch.nn as nn
from torch.nn.parameter import Parameter
import torch
import torch.nn.functional as F
import math
import mpo
import numpy as np

class Linear(nn.Module):
    def __init__(self, in_feat, out_feat,bias=True):
        super(Linear,self).__init__()
        self.in_feat=in_feat
        self.out_feat=out_feat
        self.weight=Parameter(torch.Tensor(out_feat,in_feat))
        if bias:
           self.bias=Parameter(torch.Tensor(out_feat))
        self.reset_parameters()
    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            torch.nn.init.uniform_(self.bias, -bound, bound)
    def forward(self,input):
        return F.linear(input, self.weight, self.bias)

class PoetryModel(nn.Module):
    # def __init__(self, args, vocab_size, embedding_dim, hidden_dim=1024, mpo=True):
    def __init__(self, args):
        super(PoetryModel, self).__init__()
        self.args = args
        self.hidden_dim = args.hidden_dim
        self.embeddings = nn.Embedding(args.vocab_size, args.embedding_dim)
        self.gru = nn.GRU(args.embedding_dim, self.hidden_dim, num_layers=args.num_layers)
        self.bond_dim = args.chi
        self.dims = []
        for dim in args.dims.split(';'):
            self.dims.append([int(i) for i in dim.split(",")])
        assert (np.prod(self.dims[0]) == self.hidden_dim)
        assert (np.prod(self.dims[-1]) == args.vocab_size)
        if args.mpo == True and args.vocab_size == 8400:
            if args.net == 'mpo':
                self.linear = mpo.MPO(self.hidden_dim, args.vocab_size, self.dims[0], self.dims[1], self.bond_dim)
            elif args.net == 'mpo2':
                self.linear = mpo.MPO2(self.dims[0], self.dims[1], chi=args.chi)
            elif args.net == 'mpo3':
                self.linear = mpo.MPO3(self.dims[0], self.dims[-1], chi=args.chi, chi2=args.chi2, mpo3layer=args.mpo3layers)
            elif args.net == 'operator':
                self.linear = mpo.Operator(self.dims[0], self.dims[1])
            elif args.net == 'operator2':
                self.linear = mpo.Operator2(self.dims[0], self.dims[1], chi=args.chi)
        elif args.mpo == False:
            self.linear = Linear(self.hidden_dim, args.vocab_size)

    def forward(self, input, hidden=None):
        seq_len, batch_size = input.size()
        #print(input.shape)
        if hidden is None:
            h_0 = input.data.new(args.num_layers, batch_size, self.hidden_dim).fill_(0).float()
            #c_0 = input.data.new(Config.num_layers, batch_size, self.hidden_dim).fill_(0).float()
        else:
            #h_0, c_0 = hidden
            h_0 = hidden
        embeds = self.embeddings(input)
        #output, hidden = self.lstm(embeds, (h_0, c_0))
        output, hidden = self.gru(embeds, h_0)
        output = self.linear(output.view(seq_len * batch_size, -1))
        #output = self.linear(output)
        return output, hidden
